# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants shared for the Linking API usage for the Meridian UI.

Defines constants used in the URL generation process, such as API endpoints and
parameter names.
"""

REPORT_TEMPLATE_ID = 'fbd3aeff-fc00-45fd-83f7-1ec5f21c9f56'
COMMUNITY_CONNECTOR_NAME = 'community'
COMMUNITY_CONNECTOR_ID = (
    'AKfycbz-xdEN-GbTuQ9MjEddS-64wLgXwMMTp9a4zFE4PO_kwT6wDgZPsN4Y19oKmLLHD6xk'
)
SHEETS_CONNECTOR_NAME = 'googleSheets'
GA4_MEASUREMENT_ID = 'G-R6C81BNHJ4'
